!-------------------------------------------------------------------------------

!                      Code_Saturne version 2.0.0-rc2
!                      --------------------------

!     This file is part of the Code_Saturne Kernel, element of the
!     Code_Saturne CFD tool.

!     Copyright (C) 1998-2009 EDF S.A., France

!     contact: saturne-support@edf.fr

!     The Code_Saturne Kernel is free software; you can redistribute it
!     and/or modify it under the terms of the GNU General Public License
!     as published by the Free Software Foundation; either version 2 of
!     the License, or (at your option) any later version.

!     The Code_Saturne Kernel is distributed in the hope that it will be
!     useful, but WITHOUT ANY WARRANTY; without even the implied warranty
!     of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!     GNU General Public License for more details.

!     You should have received a copy of the GNU General Public License
!     along with the Code_Saturne Kernel; if not, write to the
!     Free Software Foundation, Inc.,
!     51 Franklin St, Fifth Floor,
!     Boston, MA  02110-1301  USA

!-------------------------------------------------------------------------------

subroutine usproj &
!================

 ( idbia0 , idbra0 ,                                              &
   ndim   , ncelet , ncel   , nfac   , nfabor , nfml   , nprfml , &
   nnod   , lndfac , lndfbr , ncelbr ,                            &
   nvar   , nscal  , nphas  ,                                     &
   nbpmax , nvp    , nvep   , nivep  , ntersl , nvlsta , nvisbr , &
   nideve , nrdeve , nituse , nrtuse ,                            &
   ifacel , ifabor , ifmfbr , ifmcel , iprfml , maxelt , lstelt , &
   ipnfac , nodfac , ipnfbr , nodfbr , itepa  ,                   &
   idevel , ituser , ia     ,                                     &
   xyzcen , surfac , surfbo , cdgfac , cdgfbo , xyznod , volume , &
   dt     , rtpa   , rtp    , propce , propfa , propfb ,          &
   coefa  , coefb  ,                                              &
   ettp   , ettpa  , tepa   , statis , stativ , tslagr , parbor , &
   rdevel , rtuser , ra     )

!===============================================================================
! Purpose:
! -------

!    User subroutine.

!    Called at end of each time step, very general purpose
!    (i.e. anything that does not have another dedicated user subroutine)


! Several examples are given here:

!  - compute a thermal balance
!    (if needed, see note  below on adapting this to any scalar)

!  - compute global efforts on a subset of faces

!  - arbitrarily modify a calculation variable

!  - extract a 1 d profile

!  - print a moment

!  - examples on using parallel utility functions

! These examples are valid when using periodicity (iperio .gt. 0)
! and in parallel (irangp .ge. 0).

! The thermal balance compution also illustates a few other features,
! including the required precautions in parallel or with periodicity):
! - gradient calculation
! - computation of a value depending on cells adjacent to a face
!   (see synchronization of Dt and Cp)
! - computation of a global sum in parallel (parsom)


! Cells, boundary faces and interior faces identification
! =======================================================

! Cells, boundary faces and interior faces may be identified using
! the subroutines 'getcel', 'getfbr' and 'getfac' (respectively).

!  getfbr(string, nelts, eltlst):
!  - string is a user-supplied character string containing selection criteria;
!  - nelts is set by the subroutine. It is an integer value corresponding to
!    the number of boundary faces verifying the selection criteria;
!  - lstelt is set by the subroutine. It is an integer array of size nelts
!    containing the list of boundary faces verifying the selection criteria.

!  string may contain:
!  - references to colors (ex.: 1, 8, 26, ...)
!  - references to groups (ex.: inlet, group1, ...)
!  - geometric criteria (ex. x < 0.1, y >= 0.25, ...)
!  These criteria may be combined using logical operators ('and', 'or') and
!  parentheses.
!  Example: '1 and (group2 or group3) and y < 1' will select boundary faces
!  of color 1, belonging to groups 'group2' or 'group3' and with face center
!  coordinate y less than 1.

! Similarly, interior faces and cells can be identified using the 'getfac'
! and 'getcel' subroutines (respectively). Their syntax are identical to
! 'getfbr' syntax.

! For a more thorough description of the criteria syntax, it can be referred
! to the user guide.


!-------------------------------------------------------------------------------
! Arguments
!__________________.____._____.________________________________________________.
! name             !type!mode ! role                                           !
!__________________!____!_____!________________________________________________!
! idbia0           ! i  ! <-- ! number of first free position in ia            !
! idbra0           ! i  ! <-- ! number of first free position in ra            !
! ndim             ! i  ! <-- ! spatial dimension                              !
! ncelet           ! i  ! <-- ! number of extended (real + ghost) cells        !
! ncel             ! i  ! <-- ! number of cells                                !
! nfac             ! i  ! <-- ! number of interior faces                       !
! nfabor           ! i  ! <-- ! number of boundary faces                       !
! nfml             ! i  ! <-- ! number of families (group classes)             !
! nprfml           ! i  ! <-- ! number of properties per family (group class)  !
! nnod             ! i  ! <-- ! number of vertices                             !
! lndfac           ! i  ! <-- ! size of nodfac indexed array                   !
! lndfbr           ! i  ! <-- ! size of nodfbr indexed array                   !
! ncelbr           ! i  ! <-- ! number of cells with faces on boundary         !
! nvar             ! i  ! <-- ! total number of variables                      !
! nscal            ! i  ! <-- ! total number of scalars                        !
! nphas            ! i  ! <-- ! number of phases                               !
! nbpmax           ! i  ! <-- ! max. number of particles allowed               !
! nvp              ! i  ! <-- ! number of particle-defined variables           !
! nvep             ! i  ! <-- ! number of real particle properties             !
! nivep            ! i  ! <-- ! number of integer particle properties          !
! ntersl           ! i  ! <-- ! number of return coupling source terms         !
! nvlsta           ! i  ! <-- ! number of Lagrangian statistical variables     !
! nvisbr           ! i  ! <-- ! number of boundary statistics                  !
! nideve, nrdeve   ! i  ! <-- ! sizes of idevel and rdevel arrays              !
! nituse, nrtuse   ! i  ! <-- ! sizes of ituser and rtuser arrays              !
! ifacel(2, nfac)  ! ia ! <-- ! interior faces -> cells connectivity           !
! ifabor(nfabor)   ! ia ! <-- ! boundary faces -> cells connectivity           !
! ifmfbr(nfabor)   ! ia ! <-- ! boundary face family numbers                   !
! ifmcel(ncelet)   ! ia ! <-- ! cell family numbers                            !
! iprfml           ! ia ! <-- ! property numbers per family                    !
!  (nfml, nprfml)  !    !     !                                                !
! maxelt           ! i  ! <-- ! max number of cells and faces (int/boundary)   !
! lstelt(maxelt)   ! ia ! --- ! work array                                     !
! ipnfac(nfac+1)   ! ia ! <-- ! interior faces -> vertices index (optional)    !
! nodfac(lndfac)   ! ia ! <-- ! interior faces -> vertices list (optional)     !
! ipnfbr(nfabor+1) ! ia ! <-- ! boundary faces -> vertices index (optional)    !
! nodfbr(lndfbr)   ! ia ! <-- ! boundary faces -> vertices list (optional)     !
! itepa            ! ia ! <-- ! integer particle attributes                    !
!  (nbpmax, nivep) !    !     !   (containing cell, ...)                       !
! idevel(nideve)   ! ia ! <-- ! integer work array for temporary development   !
! ituser(nituse)   ! ia ! <-- ! user-reserved integer work array               !
! ia(*)            ! ia ! --- ! main integer work array                        !
! xyzcen           ! ra ! <-- ! cell centers                                   !
!  (ndim, ncelet)  !    !     !                                                !
! surfac           ! ra ! <-- ! interior faces surface vectors                 !
!  (ndim, nfac)    !    !     !                                                !
! surfbo           ! ra ! <-- ! boundary faces surface vectors                 !
!  (ndim, nfabor)  !    !     !                                                !
! cdgfac           ! ra ! <-- ! interior faces centers of gravity              !
!  (ndim, nfac)    !    !     !                                                !
! cdgfbo           ! ra ! <-- ! boundary faces centers of gravity              !
!  (ndim, nfabor)  !    !     !                                                !
! xyznod           ! ra ! <-- ! vertex coordinates (optional)                  !
!  (ndim, nnod)    !    !     !                                                !
! volume(ncelet)   ! ra ! <-- ! cell volumes                                   !
! dt(ncelet)       ! ra ! <-- ! time step (per cell)                           !
! rtp, rtpa        ! ra ! <-- ! calculated variables at cell centers           !
!  (ncelet, *)     !    !     !  (at current and previous time steps)          !
! propce(ncelet, *)! ra ! <-- ! physical properties at cell centers            !
! propfa(nfac, *)  ! ra ! <-- ! physical properties at interior face centers   !
! propfb(nfabor, *)! ra ! <-- ! physical properties at boundary face centers   !
! coefa, coefb     ! ra ! <-- ! boundary conditions                            !
!  (nfabor, *)     !    !     !                                                !
! ettp, ettpa      ! ra ! <-- ! particle-defined variables                     !
!  (nbpmax, nvp)   !    !     !  (at current and previous time steps)          !
! tepa             ! ra ! <-- ! real particle properties                       !
!  (nbpmax, nvep)  !    !     !  (statistical weight, ...                      !
! statis           ! ra ! <-- ! statistic means                                !
!  (ncelet, nvlsta)!    !     !                                                !
! stativ(ncelet,   ! ra ! <-- ! accumulator for variance of volume statisitics !
!        nvlsta -1)!    !     !                                                !
! tslagr           ! ra ! <-- ! Lagrangian return coupling term                !
!  (ncelet, ntersl)!    !     !  on carrier phase                              !
! parbor           ! ra ! <-- ! particle interaction properties                !
!  (nfabor, nvisbr)!    !     !  on boundary faces                             !
! rdevel(nrdeve)   ! ra ! <-> ! real work array for temporary development      !
! rtuser(nrtuse)   ! ra ! <-- ! user-reserved real work array                  !
! ra(*)            ! ra ! --- ! main real work array                           !
!__________________!____!_____!________________________________________________!

!     Type: i (integer), r (real), s (string), a (array), l (logical),
!           and composite types (ex: ra real array)
!     mode: <-- input, --> output, <-> modifies data, --- work array
!===============================================================================

implicit none

!===============================================================================
! Common blocks
!===============================================================================

include "dimfbr.h"
include "paramx.h"
include "pointe.h"
include "numvar.h"
include "optcal.h"
include "cstphy.h"
include "cstnum.h"
include "entsor.h"
include "lagpar.h"
include "lagran.h"
include "parall.h"
include "period.h"
include "ppppar.h"
include "ppthch.h"
include "ppincl.h"

!===============================================================================

! Arguments

integer          idbia0 , idbra0
integer          ndim   , ncelet , ncel   , nfac   , nfabor
integer          nfml   , nprfml
integer          nnod   , lndfac , lndfbr , ncelbr
integer          nvar   , nscal  , nphas
integer          nbpmax , nvp    , nvep  , nivep
integer          ntersl , nvlsta , nvisbr
integer          nideve , nrdeve , nituse , nrtuse

integer          ifacel(2,nfac) , ifabor(nfabor)
integer          ifmfbr(nfabor) , ifmcel(ncelet)
integer          iprfml(nfml,nprfml)
integer          maxelt, lstelt(maxelt)
integer          ipnfac(nfac+1), nodfac(lndfac)
integer          ipnfbr(nfabor+1), nodfbr(lndfbr)
integer          itepa(nbpmax,nivep)
integer          idevel(nideve), ituser(nituse)
integer          ia(*)

double precision xyzcen(ndim,ncelet)
double precision surfac(ndim,nfac), surfbo(ndim,nfabor)
double precision cdgfac(ndim,nfac), cdgfbo(ndim,nfabor)
double precision xyznod(ndim,nnod), volume(ncelet)
double precision dt(ncelet), rtp(ncelet,*), rtpa(ncelet,*)
double precision propce(ncelet,*)
double precision propfa(nfac,*), propfb(ndimfb,*)
double precision coefa(ndimfb,*), coefb(ndimfb,*)
double precision ettp(nbpmax,nvp) , ettpa(nbpmax,nvp)
double precision tepa(nbpmax,nvep)
double precision statis(ncelet,nvlsta), stativ(ncelet,nvlsta-1)
double precision tslagr(ncelet,ntersl)
double precision parbor(nfabor,nvisbr)
double precision rdevel(nrdeve), rtuser(nrtuse), ra(*)

! Local variables

integer          idebia, idebra
integer          iel    , ielg   , ifac   , ifacg  , ivar
integer          iel1   , iel2   , ieltsm
integer          iortho , impout
integer          ifinia , ifinra
integer          igradx , igrady , igradz
integer          itravx , itravy , itravz , itreco
integer          inc    , iccocg
integer          nswrgp , imligp , iphydp , iwarnp
integer          iutile , iphas  , iclvar , iii
integer          ipcrom , ipcvst , iflmas , iflmab , ipccp, ipcvsl
integer          idimte , itenso , iscal
integer          ii     , nbr    , irangv , irang1 , npoint
integer          imom   , ipcmom , idtcm
integer          itab(3), iun
integer          ncesmp , icesmp , ismacp , itpsmp
integer          ilelt  , nlelt

double precision xrtpa  , xrtp
double precision xbilan , xbilvl , xbilpa , xbilpt
double precision xbilsy , xbilen , xbilso , xbildv
double precision xbilmi , xbilma
double precision epsrgp , climgp , extrap
double precision xfluxf , xgamma
double precision diipbx, diipby, diipbz, surfbn, distbr
double precision visct, flumab , xcp , xvsl, cp0iph, rrr
double precision xfor(3), xyz(3), xabs, xu, xv, xw, xk, xeps

!james
!profiles
integer          jj, kk
double precision xpoint(4), xmom(4), cp
character        xname(4)

!gradient

integer          bfile
double precision xyzx, xyzy, xyzz
double precision dpdx, dpdy, dpdz
double precision iprx
double precision iux,  iuy,  iuz

!===============================================================================


!===============================================================================
! 1.  Initialization
!===============================================================================

! Memory management

idebia = idbia0
idebra = idbra0

!===============================================================================

!===============================================================================
! 5. Example: extraction of a 1D profile
!    -----------------------------------

! We seek here to extract the profile of U, V, W, k and epsilon on an
! arbitrary 1D curve based on a curvilear abscissa.
! The profile is described in the 'profile.dat' file (do not forget to
! define it as user data in the run script).

! - the curve used here is the segment: [(0;0;0),(0;0.1;0)], but the
!   generalization to an arbitrary curve is simple.
! - the routine handles parallelism an periodicity, as well as the different
!   turbulence models.
! - the 1D curve is discretized into 'npoint' points. For each of these
!   points, we search for the closest cell center and we output the variable
!   values at this cell center. For better consistency, the coordinate
!   which is output is that of the cell center (instead of the initial point).
! - we avoid using the same cell multiple times (in case several points
!   an the curve are associated with the same cell).
!===============================================================================

! ----------------------------------------------

xpoint(1) = -1.d1
xpoint(2) = -1.d-3
xpoint(3) =  1.d-3
xpoint(4) =  1.d1

xname(1) = '1'
xname(2) = '2'
xname(3) = '3'
xname(4) = '4'


! ----------------------------------------------

if (ntcabs.eq.ntmabs) then
  do jj=1,2
    ! Only process of rank 0 (parallel) or -1 (scalar) writes to this file.
    ! We use 'user' Fortran units.
    impout = impusr(1)
    if (irangp.le.0) then
      open(impout,file='profile.'//xname(jj)//'.dat')
      write(impout,*)  &
           '# z(m) U(m/s) V(m/s) W(m/s) Cp'
    endif

    iphas  = 1
    npoint = 200
    iel1   = -999
    irang1 = -999
    do ii = 1, npoint

      xyz(1) = xpoint(jj)
      xyz(2) = 0.0d0
      xyz(3) = float(ii-1)/float(npoint-1)*0.5d0 - 0.25d0
      call findpt(ncelet, ncel, xyzcen, xyz(1), xyz(2), xyz(3), iel, irangv)
      !==========
      
      if ((iel.ne.iel1).or.(irangv.ne.irang1)) then
        iel1   = iel
        irang1 = irangv
        ! Set temporary variables xu, xv, ... for the process containing
        ! the point and then send it to other processes.
        do kk=1, 4

          imom = kk ! Moment number

          ! Position in 'propce' of the array of temporal accumulation for moments,
          ! propce(iel,ipcmom)

          ipcmom = ipproc(icmome(imom))

          ! The temporal accumulation for moments must be divided by the accumulated
          ! time, which id an array of size ncel or a single real number:
          ! - array of size ncel if idtmom(imom) > 0 : propce(iel, idtcm)
          ! - or simple real     if idtmom(imom) < 0 : dtcmom(idtcm)

          if (idtmom(imom).gt.0) then
            idtcm = ipproc(icdtmo(idtmom(imom)))
            xmom(kk) = propce(iel, ipcmom)/max(propce(iel, idtcm), epzero)
          elseif (idtmom(imom).lt.0) then
            idtcm = -idtmom(imom)
            xmom(kk) = propce(iel, ipcmom)/max(dtcmom(idtcm), epzero)
          endif
        enddo

        if (irangp.eq.irangv) then
          xabs = xyzcen(3,iel)
          xu   = xmom(1)
          xv   = xmom(2)
          xw   = xmom(3)
          cp   = xmom(4)
        else
          xabs = 0.d0
          xu   = 0.d0
          xv   = 0.d0
          xw   = 0.d0
          cp   = 0.d0
        endif
        cp = cp / 0.5 / ro0(iphas) / uref(iphas) / uref(iphas)

      ! Broadcast to other ranks in parallel
        if (irangp.ge.0) then
          iun = 1
          call parbcr(irangv, iun, xabs)
          call parbcr(irangv, iun, xu)
          call parbcr(irangv, iun, xv)
          call parbcr(irangv, iun, xw)
          call parbcr(irangv, iun, cp)
        endif

        if (irangp.le.0) write(impout,99) xabs, xu, xv, xw, cp


      endif

    enddo

    if (irangp.le.0) close(impout)
  enddo

! gradient reconstruction

! Reserve memory
  ifinia = idebia

  igradx = idebra
  igrady = igradx + ncelet
  igradz = igrady + ncelet
  itravx = igradz + ncelet
  itravy = itravx + ncelet
  itravz = itravy + ncelet
  ifinra = itravz + nfabor

! Check memory availability

  call iasize('usproj', ifinia)
  call rasize('usproj', ifinra)

! define variable and BC number
  iphas  = 1
  ivar   = ipr(iphas)
  iclvar = iclrtp(ivar,icoef)

! - Parallel update
  if (irangp.ge.0) then
    call parcom (rtp(1,ivar))
    !    ======
  endif

  ! - Periodic update
  if (iperio.eq.1) then

    idimte = 0
    itenso = 0
    call percom                                 &
    !    ======
      ( idimte , itenso ,                       &
        rtp(1,ivar), rtp(1,ivar), rtp(1,ivar),  &
        rtp(1,ivar), rtp(1,ivar), rtp(1,ivar),  &
        rtp(1,ivar), rtp(1,ivar), rtp(1,ivar))

  endif


! - Compute gradient

  inc = 1
  iccocg = 1
  nswrgp = nswrgr(ivar)
  imligp = imligr(ivar)
  iwarnp = iwarni(ivar)
  epsrgp = epsrgr(ivar)
  climgp = climgr(ivar)
  extrap = extrag(ivar)
  iphydp = 0

  call grdcel                                                     &
  !    ======
    ( ifinia , ifinra ,                                              &
      ndim   , ncelet , ncel   , nfac   , nfabor , nfml   , nprfml , &
      nnod   , lndfac , lndfbr , ncelbr , nphas  ,                   &
      nideve , nrdeve , nituse , nrtuse ,                            &
      ivar   , imrgra , inc    , iccocg , nswrgp , imligp , iphydp , &
      iwarnp , nfecra ,                                              &
      epsrgp , climgp , extrap ,                                     &
      ifacel , ifabor , ifmfbr , ifmcel , iprfml ,                   &
      ipnfac , nodfac , ipnfbr , nodfbr ,                            &
      idevel , ituser , ia     ,                                     &
      xyzcen , surfac , surfbo , cdgfac , cdgfbo , xyznod , volume , &
      ra(itravx) , ra(itravx) , ra(itravx) ,                         &
      rtp(1,ivar) , coefa(1,iclvar) , coefb(1,iclvar) ,              &
      ra(igradx) , ra(igrady) , ra(igradz) ,                         &
      !---------   ----------   ----------
      ra(itravx) , ra(itravy) , ra(itravz) ,                         &
      rdevel , rtuser , ra     )

! - list values in boundary cells

  bfile = impusr(2)
  open(bfile,file='interface_left.dat')

  call getfbr('interface', nlelt, lstelt)
  !    ======

  do ilelt = 1, nlelt
  
    ifac   = lstelt(ilelt)
    iel    = ifabor(ifac)

    !pressure gradients
    dpdx   = ra(igradx+iel-1)
    dpdy   = ra(igrady+iel-1)
    dpdz   = ra(igradz+iel-1)
  
    xyzx   = xyzcen(1,iel)
    xyzy   = xyzcen(2,iel)
    xyzz   = xyzcen(3,iel)
  
    iprx   = rtp(iel,ipr(iphas))

    iux    = rtp(iel, iu(iphas))
    iuy    = rtp(iel, iv(iphas))
    iuz    = rtp(iel, iw(iphas))

    
    write(bfile,80) xyzx, xyzy, xyzz, &
                    iux,  iuy,  iuz,  &
                    iprx,             &
                    dpdx, dpdy, dpdz

  enddo

  close(bfile)

endif

!Formats
80    format(10e17.9)
99    format(5e17.9)

return
end subroutine
